"""Generate the Python data structures from the intermediate representation."""
import io
import itertools
import textwrap
from typing import (
    Optional,
    Dict,
    List,
    Tuple,
    cast,
    Union,
    Mapping,
    Final,
)

from icontract import ensure, require

from aas_core_codegen import intermediate
from aas_core_codegen import specific_implementations
from aas_core_codegen.common import (
    Error,
    Identifier,
    assert_never,
    Stripped,
    indent_but_first_line,
)
from aas_core_codegen.intermediate import (
    construction as intermediate_construction,
)
from aas_core_codegen.python import (
    common as python_common,
    naming as python_naming,
    description as python_description,
    unrolling as python_unrolling,
)
from aas_core_codegen.python.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
)


# region Checks


def _human_readable_identifier(
    something: Union[
        intermediate.Enumeration, intermediate.AbstractClass, intermediate.ConcreteClass
    ]
) -> str:
    """
    Represent ``something`` in a human-readable text.

    The reader should be able to trace ``something`` back to the meta-model.
    """
    result: str

    if isinstance(something, intermediate.Enumeration):
        result = f"meta-model enumeration {something.name!r}"
    elif isinstance(something, intermediate.AbstractClass):
        result = f"meta-model abstract class {something.name!r}"
    elif isinstance(something, intermediate.ConcreteClass):
        result = f"meta-model concrete class {something.name!r}"
    else:
        assert_never(something)

    return result


def _verify_intra_structure_collisions(
    our_type: intermediate.OurType,
) -> Optional[Error]:
    """Verify that no member names collide in the Python structure of our type."""
    errors = []  # type: List[Error]

    if isinstance(our_type, intermediate.Enumeration):
        enum_literal_map = (
            dict()
        )  # type: Dict[Identifier, intermediate.EnumerationLiteral]

        for literal in our_type.literals:
            literal_name = python_naming.enum_literal_name(literal.name)
            colliding_literal = enum_literal_map.get(literal_name, None)
            if colliding_literal is not None:
                errors.append(
                    Error(
                        literal.parsed.node,
                        f"The Python name, {literal_name!r}, "
                        f"for the literal {literal.name!r} collides with "
                        f"the Python name of another "
                        f"literal {colliding_literal.name!r}",
                    )
                )
            else:
                enum_literal_map[literal_name] = literal

    elif isinstance(our_type, intermediate.ConstrainedPrimitive):
        pass

    elif isinstance(our_type, intermediate.Class):
        observed_member_names = {}  # type: Dict[Identifier, str]

        for prop in our_type.properties:
            prop_name = python_naming.property_name(prop.name)
            if prop_name in observed_member_names:
                errors.append(
                    Error(
                        prop.parsed.node,
                        f"Python property {prop_name!r} corresponding "
                        f"to the meta-model property {prop.name!r} collides with "
                        f"the {observed_member_names[prop_name]}",
                    )
                )
            else:
                observed_member_names[prop_name] = (
                    f"Python property {prop_name!r} corresponding to "
                    f"the meta-model property {prop.name!r}"
                )

        for method in our_type.methods:
            method_name = python_naming.method_name(method.name)

            if method_name in observed_member_names:
                errors.append(
                    Error(
                        method.parsed.node,
                        f"Python method {method_name!r} corresponding "
                        f"to the meta-model method {method.name!r} collides with "
                        f"the {observed_member_names[method_name]}",
                    )
                )
            else:
                observed_member_names[method_name] = (
                    f"Python method {method_name!r} corresponding to "
                    f"the meta-model method {method.name!r}"
                )

    else:
        assert_never(our_type)

    if len(errors) > 0:
        errors.append(
            Error(
                our_type.parsed.node,
                f"Naming collision(s) in Python code for our type {our_type.name!r}",
                underlying=errors,
            )
        )

    return None


def _verify_structure_name_collisions(
    symbol_table: intermediate.SymbolTable,
) -> List[Error]:
    """Verify that the Python names of the structures do not collide."""
    observed_structure_names: Dict[
        Identifier,
        Union[
            intermediate.Enumeration,
            intermediate.AbstractClass,
            intermediate.ConcreteClass,
        ],
    ] = dict()

    errors = []  # type: List[Error]

    # region Inter-structure collisions

    for enum_or_cls in itertools.chain(symbol_table.enumerations, symbol_table.classes):
        name = python_naming.name_of(enum_or_cls)

        other = observed_structure_names.get(name, None)

        if other is not None:
            errors.append(
                Error(
                    enum_or_cls.parsed.node,
                    f"The Python name {name!r} "
                    f"of the {_human_readable_identifier(enum_or_cls)} "
                    f"collides with the Python name "
                    f"of the {_human_readable_identifier(other)}",
                )
            )
        else:
            observed_structure_names[name] = enum_or_cls

    # endregion

    # region Intra-structure collisions

    for our_type in symbol_table.our_types:
        collision_error = _verify_intra_structure_collisions(our_type=our_type)

        if collision_error is not None:
            errors.append(collision_error)

    # endregion

    return errors


class VerifiedIntermediateSymbolTable(intermediate.SymbolTable):
    """Represent a verified symbol table which can be used for code generation."""

    # noinspection PyInitNewSignature
    def __new__(
        cls, symbol_table: intermediate.SymbolTable
    ) -> "VerifiedIntermediateSymbolTable":
        raise AssertionError("Only for type annotation")


@ensure(lambda result: (result[0] is None) ^ (result[1] is None))
def verify(
    symbol_table: intermediate.SymbolTable,
) -> Tuple[Optional[VerifiedIntermediateSymbolTable], Optional[List[Error]]]:
    """Verify that Python code can be generated from the ``symbol_table``."""
    errors = []  # type: List[Error]

    structure_name_collisions = _verify_structure_name_collisions(
        symbol_table=symbol_table
    )

    errors.extend(structure_name_collisions)

    if len(errors) > 0:
        return None, errors

    return cast(VerifiedIntermediateSymbolTable, symbol_table), None


# endregion

# region Generation


@require(lambda enumeration, literal: id(literal) in enumeration.literal_id_set)
@require(lambda literal: literal.description is not None)
def _generate_comment_for_enumeration_literal(
    enumeration: intermediate.Enumeration,
    literal: intermediate.EnumerationLiteral,
    qualified_module_name: python_common.QualifiedModuleName,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the documentation comment for the given enumeration literal."""
    # NOTE (mristin, 2022-10-29):
    # We need to state the pre-condition for the second time for mypy.
    assert literal.description is not None

    text, errors = python_description.generate_summary_remarks(
        description=literal.description,
        context=python_description.Context(
            qualified_module_name=qualified_module_name,
            module=Identifier("types"),
            cls_or_enum=enumeration,
        ),
    )

    if errors is not None:
        return None, errors

    assert text is not None

    return python_description.documentation_comment(text), None


@require(lambda cls_or_enum: cls_or_enum.description is not None)
def _generate_docstring_for_cls_or_enum(
    cls_or_enum: Union[intermediate.Enumeration, intermediate.ClassUnion],
    qualified_module_name: python_common.QualifiedModuleName,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the docstring for our type."""
    # NOTE (mristin, 2022-10-29):
    # We need to state the pre-condition for the second time for mypy.
    assert cls_or_enum.description is not None

    text, errors = python_description.generate_summary_remarks_constraints(
        description=cls_or_enum.description,
        context=python_description.Context(
            qualified_module_name=qualified_module_name,
            module=Identifier("types"),
            cls_or_enum=cls_or_enum,
        ),
    )

    if errors is not None:
        return None, errors

    assert text is not None

    return python_description.docstring(text), None


@ensure(lambda result: (result[0] is None) ^ (result[1] is None))
def _generate_enum(
    enum: intermediate.Enumeration,
    qualified_module_name: python_common.QualifiedModuleName,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the Python code for the enum."""
    writer = io.StringIO()

    errors = []  # type: List[Error]

    docstring = None  # type: Optional[Stripped]
    if enum.description is not None:
        # fmt: off
        docstring, docstring_errors = (
            _generate_docstring_for_cls_or_enum(
                cls_or_enum=enum,
                qualified_module_name=qualified_module_name
            )
        )
        # fmt: on

        if docstring_errors:
            errors.append(
                Error(
                    enum.description.parsed.node,
                    f"Failed to generate the docstring "
                    f"for the enumeration {enum.name!r}",
                    docstring_errors,
                )
            )
        else:
            assert docstring is not None

    name = python_naming.enum_name(enum.name)

    writer.write(f"class {name}(enum.Enum):\n")
    if len(enum.literals) == 0:
        if docstring is not None:
            writer.write(textwrap.indent(docstring, I))
        else:
            writer.write(
                f"""\
{I}# pylint: disable=missing-class-docstring
{I}pass
"""
            )
    else:
        if docstring is not None:
            writer.write(textwrap.indent(docstring, I))
        else:
            writer.write(f"{I}# pylint: disable=missing-class-docstring")

        for literal in enum.literals:
            writer.write("\n\n")

            if literal.description is not None:
                comment, comment_errors = _generate_comment_for_enumeration_literal(
                    enumeration=enum,
                    literal=literal,
                    qualified_module_name=qualified_module_name,
                )
                if comment_errors is not None:
                    errors.append(
                        Error(
                            literal.description.parsed.node,
                            f"Failed to generate the documentation comment "
                            f"for enumeration literal {literal.name!r}",
                            comment_errors,
                        )
                    )
                else:
                    assert comment is not None

                    writer.write(textwrap.indent(comment, I))
                    writer.write("\n")

            literal_name = python_naming.enum_literal_name(literal.name)
            writer.write(textwrap.indent(f"{literal_name} = {repr(literal.value)}", I))

    if len(errors) > 0:
        return None, Error(
            enum.parsed.node,
            f"Failed to generate the Python code for the enumeration {enum.name!r}",
            errors,
        )

    return Stripped(writer.getvalue()), None


class _DescendBodyUnroller(python_unrolling.AbstractUnroller):
    """Generate the code that unrolls descent into an element."""

    #: If set, generates the code with unrolled yields.
    #: Otherwise, we do not unroll recursively.
    _recurse: Final[bool]

    #: Pre-computed descendability map. A type is descendable if we should unroll it
    #: further.
    _descendability: Final[Mapping[intermediate.TypeAnnotationUnion, bool]]

    #: Generator of loop variable names.
    #:
    #: We generate for each list iteration a new variable since Python tracks
    #: variables in the function scope, not block scope.
    _generator_for_loop_variables: Final[python_common.GeneratorForLoopVariables]

    def __init__(
        self,
        recurse: bool,
        descendability: Mapping[intermediate.TypeAnnotationUnion, bool],
        generator_for_loop_variables: python_common.GeneratorForLoopVariables,
    ) -> None:
        """Initialize with the given values."""
        self._recurse = recurse
        self._descendability = descendability
        self._generator_for_loop_variables = generator_for_loop_variables

    def _unroll_primitive_type_annotation(
        self,
        unrollee_expr: str,
        type_annotation: intermediate.PrimitiveTypeAnnotation,
        path: List[str],
        list_loop_level: int,
    ) -> List[python_unrolling.Node]:
        """Generate code for the given specific ``type_annotation``."""
        # We can not descend into a primitive type.
        return []

    def _unroll_our_type_annotation(
        self,
        unrollee_expr: str,
        type_annotation: intermediate.OurTypeAnnotation,
        path: List[str],
        list_loop_level: int,
    ) -> List[python_unrolling.Node]:
        """Generate code for the given specific ``type_annotation``."""
        our_type = type_annotation.our_type

        if isinstance(our_type, intermediate.Enumeration):
            # We can not descend into an enumeration.
            return []

        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            # We can not descend into a primitive type.
            return []

        assert isinstance(our_type, intermediate.Class)  # Exhaustively match

        result = [python_unrolling.Node(f"yield {unrollee_expr}", children=[])]

        if self._recurse:
            if self._descendability[type_annotation]:
                result.append(
                    python_unrolling.Node(
                        text=f"yield from {unrollee_expr}.descend()",
                        children=[],
                    )
                )

        return result

    def _unroll_list_type_annotation(
        self,
        unrollee_expr: str,
        type_annotation: intermediate.ListTypeAnnotation,
        path: List[str],
        list_loop_level: int,
    ) -> List[python_unrolling.Node]:
        """Generate code for the given specific ``type_annotation``."""
        if (
            not self._recurse
            and isinstance(type_annotation.items, intermediate.OurTypeAnnotation)
            and isinstance(
                type_annotation.items.our_type,
                (intermediate.AbstractClass, intermediate.ConcreteClass),
            )
        ):
            return [python_unrolling.Node(f"yield from {unrollee_expr}", children=[])]

        loop_var = next(self._generator_for_loop_variables)
        children = self.unroll(
            unrollee_expr=loop_var,
            type_annotation=type_annotation.items,
            path=[],  # Path is unused in this context
            list_loop_level=list_loop_level + 1,
        )

        if len(children) == 0:
            return []

        node = python_unrolling.Node(
            text=f"for {loop_var} in {unrollee_expr}:",
            children=children,
        )

        return [node]

    def _unroll_optional_type_annotation(
        self,
        unrollee_expr: str,
        type_annotation: intermediate.OptionalTypeAnnotation,
        path: List[str],
        list_loop_level: int,
    ) -> List[python_unrolling.Node]:
        """Generate code for the given specific ``type_annotation``."""
        children = self.unroll(
            unrollee_expr=unrollee_expr,
            type_annotation=type_annotation.value,
            path=path,
            list_loop_level=list_loop_level,
        )

        if len(children) == 0:
            return []

        return [
            python_unrolling.Node(
                text=f"if {unrollee_expr} is not None:", children=children
            )
        ]


def _generate_descend_body(cls: intermediate.ConcreteClass, recurse: bool) -> Stripped:
    """
    Generate the body of the ``descend`` and ``descend_once`` methods.

    With this function, we unroll the recursion as a simple optimization
    in the recursive case.
    """
    blocks = []  # type: List[Stripped]

    generator_for_loop_variables = python_common.GeneratorForLoopVariables()

    for prop in cls.properties:
        descendability = intermediate.map_descendability(
            type_annotation=prop.type_annotation
        )

        if not descendability[prop.type_annotation]:
            continue

        # region Unroll

        unroller = _DescendBodyUnroller(
            recurse=recurse,
            descendability=descendability,
            generator_for_loop_variables=generator_for_loop_variables,
        )

        roots = unroller.unroll(
            unrollee_expr=f"self.{python_naming.property_name(prop.name)}",
            type_annotation=prop.type_annotation,
            path=[],  # We do not use path in this context
            list_loop_level=0,
        )

        assert len(roots) > 0, (
            "Since the type annotation was descendable, we must have obtained "
            "at least one unrolling node"
        )

        blocks.extend(Stripped(python_unrolling.render(root)) for root in roots)

        # endregion

    if len(blocks) == 0:
        blocks.append(
            Stripped(
                """\
# No descendable properties
return
# For this uncommon return-yield construction, see:
# https://stackoverflow.com/questions/13243766/how-to-define-an-empty-generator-function
# noinspection PyUnreachableCode
yield"""
            )
        )

    return Stripped("\n\n".join(blocks))


def _generate_descend_once_method(cls: intermediate.ConcreteClass) -> Stripped:
    """Generate the ``descend_once`` method for the concrete class ``cls``."""

    body = _generate_descend_body(cls=cls, recurse=False)

    return Stripped(
        f"""\
def descend_once(self) -> Iterator[Class]:
{I}\"\"\"
{I}Iterate over the instances referenced from this instance.

{I}We do not recurse into the referenced instance.

{I}:yield: instances directly referenced from this instance
{I}\"\"\"
{I}{indent_but_first_line(body, I)}"""
    )


def _generate_descend_method(cls: intermediate.ConcreteClass) -> Stripped:
    """Generate the recursive ``descend`` method for the concrete class ``cls``."""

    body = _generate_descend_body(cls=cls, recurse=True)

    return Stripped(
        f"""\
def descend(self) -> Iterator[Class]:
{I}\"\"\"
{I}Iterate recursively over the instances referenced from this one.

{I}:yield: instances recursively referenced from this instance
{I}\"\"\"
{I}{indent_but_first_line(body, I)}"""
    )


def _generate_default_value(default: intermediate.Default) -> Stripped:
    """Generate the Python code representing the default value of an argument."""
    code: str

    if isinstance(default, intermediate.DefaultPrimitive):
        if default.value is None:
            code = "None"
        elif isinstance(default.value, bool):
            code = "True" if default.value else "False"
        elif isinstance(default.value, str):
            code = python_common.string_literal(default.value)
        elif isinstance(default.value, int):
            code = str(default.value)
        elif isinstance(default.value, float):
            code = f"{default}"
        else:
            assert_never(default.value)
    elif isinstance(default, intermediate.DefaultEnumerationLiteral):
        code = ".".join(
            [
                python_naming.enum_name(default.enumeration.name),
                python_naming.enum_literal_name(default.literal.name),
            ]
        )
    else:
        assert_never(default)

    return Stripped(code)


@require(lambda cls: not cls.is_implementation_specific)
@require(lambda cls: not cls.constructor.is_implementation_specific)
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_constructor(
    cls: intermediate.ClassUnion,
    symbol_table: intermediate.SymbolTable,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """
    Generate the constructor function for the given class ``cls``.

    Return empty string if there is an empty constructor.
    """
    if len(cls.constructor.arguments) == 0 and len(cls.constructor.statements) == 0:
        return Stripped(""), None

    # region Construct the body

    body = []  # type: List[Stripped]
    for stmt in cls.constructor.statements:
        if isinstance(stmt, intermediate_construction.CallSuperConstructor):
            super_class = symbol_table.must_find_class(stmt.super_name)

            writer = io.StringIO()

            if len(super_class.constructor.arguments) == 0:
                writer.write(
                    f"""\
{python_naming.class_name(super_class.name)}.__init__(
{I}self
)"""
                )
            else:
                writer.write(
                    f"""\
{python_naming.class_name(super_class.name)}.__init__(
{I}self,
"""
                )
                for i, arg in enumerate(super_class.constructor.arguments):
                    writer.write(f"{I}{python_naming.argument_name(arg.name)}")
                    if i < len(super_class.constructor.arguments) - 1:
                        writer.write(",\n")
                    else:
                        writer.write("\n")
                writer.write(")")

            body.append(Stripped(writer.getvalue()))

        elif isinstance(stmt, intermediate_construction.AssignArgument):
            if stmt.default is None:
                body.append(
                    Stripped(
                        f"self.{python_naming.property_name(stmt.name)} = "
                        f"{python_naming.argument_name(stmt.argument)}"
                    )
                )
            else:
                if isinstance(stmt.default, intermediate_construction.EmptyList):
                    arg_name = python_naming.argument_name(stmt.argument)

                    # Write the assignment as a ternary operator
                    writer = io.StringIO()
                    writer.write(
                        f"""\
self.{python_naming.property_name(stmt.name)} = (
{I}{arg_name}
{I}if {arg_name} is not None
{I}else []
)"""
                    )

                    body.append(Stripped(writer.getvalue()))
                elif isinstance(
                    stmt.default, intermediate_construction.DefaultEnumLiteral
                ):
                    literal_code = ".".join(
                        [
                            python_naming.enum_name(stmt.default.enum.name),
                            python_naming.enum_literal_name(stmt.default.literal.name),
                        ]
                    )

                    arg_name = python_naming.argument_name(stmt.argument)

                    body.append(
                        Stripped(
                            f"""\
self.{python_naming.property_name(stmt.name)} = (
{I}{arg_name}
{I}if {arg_name} is not None
{I}else {literal_code}
)"""
                        )
                    )
                else:
                    assert_never(stmt.default)

        else:
            assert_never(stmt)

    # endregion

    # region Assemble the constructor with the definition and the body

    arg_codes = []  # type: List[str]
    for arg in cls.constructor.arguments:
        arg_type = python_common.generate_type(type_annotation=arg.type_annotation)
        arg_name = python_naming.argument_name(arg.name)

        if arg.default is None:
            arg_codes.append(Stripped(f"{arg_name}: {arg_type}"))
        else:
            arg_codes.append(
                Stripped(
                    f"{arg_name}: {arg_type} = {_generate_default_value(arg.default)}"
                )
            )

    writer = io.StringIO()

    if len(arg_codes) == 0:
        writer.write("def __init__(self) -> None:\n")
    if len(arg_codes) == 1:
        writer.write(f"def __init__(self, {arg_codes[0]}) -> None:\n")
    else:
        arg_block = ",\n".join(["self"] + arg_codes)
        arg_block_indented = textwrap.indent(arg_block, II)
        writer.write(f"def __init__(\n{arg_block_indented}\n) -> None:\n")

    writer.write(f'{I}"""Initialize with the given values."""')

    if len(body) > 0:
        for body_stmt in body:
            writer.write(f"\n{I}{indent_but_first_line(body_stmt, I)}")

    # endregion

    return Stripped(writer.getvalue()), None


@require(lambda cls, prop: id(prop) in cls.property_id_set)
@require(lambda prop: prop.description is not None)
def _generate_comment_for_property(
    cls: intermediate.ClassUnion,
    prop: intermediate.Property,
    qualified_module_name: python_common.QualifiedModuleName,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the documentation comment for the given property."""
    # NOTE (mristin, 2022-10-29):
    # We need to write a double assertion for mypy.
    assert prop.description is not None

    text, errors = python_description.generate_summary_remarks_constraints(
        description=prop.description,
        context=python_description.Context(
            qualified_module_name=qualified_module_name,
            module=Identifier("types"),
            cls_or_enum=cls,
        ),
    )

    if errors is not None:
        return None, errors

    assert text is not None

    return python_description.documentation_comment(text), None


@require(lambda cls: not cls.is_implementation_specific)
@ensure(lambda result: (result[0] is None) ^ (result[1] is None))
def _generate_class(
    cls: intermediate.ClassUnion,
    spec_impls: specific_implementations.SpecificImplementations,
    qualified_module_name: python_common.QualifiedModuleName,
    symbol_table: intermediate.SymbolTable,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate Python code for the given concrete class ``cls``."""
    # region Collect blocks of the class body

    # Code blocks of the class body separated by double newlines and indented once
    blocks = []  # type: List[Stripped]

    # region Property definitions

    for prop in cls.properties:
        if prop.specified_for is not cls:
            continue

        prop_comment = None  # type: Optional[Stripped]
        if prop.description is not None:
            prop_comment, prop_comment_errors = _generate_comment_for_property(
                cls=cls, prop=prop, qualified_module_name=qualified_module_name
            )
            if prop_comment_errors is not None:
                return None, Error(
                    prop.description.parsed.node,
                    "Failed to generate the property comment",
                    prop_comment_errors,
                )

        prop_type = python_common.generate_type(type_annotation=prop.type_annotation)
        prop_name = python_naming.property_name(prop.name)

        writer = io.StringIO()
        if prop_comment is not None:
            writer.write(prop_comment)
            writer.write("\n")
        writer.write(f"{prop_name}: {prop_type}")
        blocks.append(Stripped(writer.getvalue()))

    # endregion

    # region over_X_or_empty getter

    for prop in cls.properties:
        if prop.specified_for is not cls:
            continue

        if isinstance(
            prop.type_annotation, intermediate.OptionalTypeAnnotation
        ) and isinstance(prop.type_annotation.value, intermediate.ListTypeAnnotation):
            prop_name = python_naming.property_name(prop.name)
            items_type = python_common.generate_type(prop.type_annotation.value.items)

            blocks.append(
                Stripped(
                    f"""\
def over_{prop_name}_or_empty(
{II}self
) -> Iterator[{items_type}]:
{I}\"\"\"Yield from :py:attr:`.{prop_name}` if set.\"\"\"
{I}if self.{prop_name} is not None:
{II}yield from self.{prop_name}"""
                )
            )

    # endregion

    # region Methods

    errors = []  # type: List[Error]

    for method in cls.methods:
        if method.specified_for is not cls:
            continue

        if isinstance(method, intermediate.ImplementationSpecificMethod):
            implementation_key = specific_implementations.ImplementationKey(
                f"Types/{method.specified_for.name}/{method.name}.py"
            )

            implementation = spec_impls.get(implementation_key, None)

            if implementation is None:
                errors.append(
                    Error(
                        method.parsed.node,
                        f"The implementation is missing for "
                        f"the implementation-specific method: {implementation_key}",
                    )
                )
                continue

            blocks.append(implementation)
        else:
            # NOTE (mristin, 2022-09-28):
            # At the moment, we do not transpile the method body and its contracts.
            # We want to finish the meta-model for the V3, fix de/serialization and
            # generate SDKs for a couple of languages before taking on this rather hard
            # task.

            errors.append(
                Error(
                    cls.parsed.node,
                    "At the moment, we do not transpile the method body and "
                    "its contracts. We want to finish the meta-model for the V3, "
                    "fix de/serialization and generate SDKs for a couple of languages "
                    "before taking on this rather hard task.",
                )
            )

    if isinstance(cls, intermediate.ConcreteClass):
        blocks.append(_generate_descend_once_method(cls=cls))

        blocks.append(_generate_descend_method(cls=cls))

        visit_name = python_naming.method_name(Identifier(f"visit_{cls.name}"))

        blocks.append(
            Stripped(
                f"""\
def accept(self, visitor: "AbstractVisitor") -> None:
{I}\"\"\"Dispatch the :paramref:`visitor` on this instance.\"\"\"
{I}visitor.{visit_name}(self)"""
            )
        )

        visit_with_context_name = python_naming.method_name(
            Identifier(f"visit_{cls.name}_with_context")
        )

        blocks.append(
            Stripped(
                f"""\
def accept_with_context(
{II}self,
{II}visitor: "AbstractVisitorWithContext[ContextT]",
{II}context: ContextT
) -> None:
{I}\"\"\"Dispatch the :paramref:`visitor` on this instance in :paramref:`context`.\"\"\"
{I}visitor.{visit_with_context_name}(self, context)"""
            )
        )

        transform_name = python_naming.method_name(Identifier(f"transform_{cls.name}"))

        blocks.append(
            Stripped(
                f"""\
def transform(
{II}self,
{II}transformer: "AbstractTransformer[T]"
) -> T:
{I}\"\"\"Dispatch the :paramref:`transformer` on this instance.\"\"\"
{I}return transformer.{transform_name}(self)"""
            )
        )

        transform_with_context_name = python_naming.method_name(
            Identifier(f"transform_{cls.name}_with_context")
        )

        blocks.append(
            Stripped(
                f"""\
def transform_with_context(
{II}self,
{II}transformer: "AbstractTransformerWithContext[ContextT, T]",
{II}context: ContextT
) -> T:
{I}\"\"\"
{I}Dispatch the :paramref:`transformer` on this instance in :paramref:`context`.
{I}\"\"\"
{I}return transformer.{transform_with_context_name}(
{II}self, context)"""
            )
        )

    # endregion

    # region Constructor

    if cls.constructor.is_implementation_specific:
        implementation_key = specific_implementations.ImplementationKey(
            f"Types/{cls.name}/{cls.name}.py"
        )
        implementation = spec_impls.get(implementation_key, None)

        if implementation is None:
            errors.append(
                Error(
                    cls.parsed.node,
                    f"The implementation of the implementation-specific constructor "
                    f"is missing: {implementation_key}",
                )
            )
        else:
            blocks.append(implementation)
    else:
        constructor_block, error = _generate_constructor(
            cls=cls, symbol_table=symbol_table
        )

        if error is not None:
            errors.append(error)
        else:
            assert constructor_block is not None

            # NOTE (mristin, 2022-06-21):
            # Empty constructor will be automatically generated by the interpreter.
            if constructor_block != "":
                blocks.append(constructor_block)

    # endregion

    if len(errors) > 0:
        return None, Error(
            cls.parsed.node,
            f"Failed to generate the code for the class {cls.name}",
            errors,
        )

    docstring = None  # type: Optional[Stripped]
    if cls.description is not None:
        docstring, docstring_errors = _generate_docstring_for_cls_or_enum(
            cls_or_enum=cls, qualified_module_name=qualified_module_name
        )
        if docstring_errors is not None:
            return None, Error(
                cls.description.parsed.node,
                "Failed to generate the docstring",
                docstring_errors,
            )

        assert docstring is not None

    # endregion

    # region Assemble the class definition with its body

    name = python_naming.class_name(cls.name)

    # NOTE (mristin, 2022-09-17):
    # Every class in the meta-model inherits from the most general abstract class
    # ``Class`.

    super_names = []  # type: List[str]

    if len(cls.inheritances) == 0:
        # NOTE (mristin, 2022-09-17):
        # We only have to include the most general ancestor ``Class`` if there are no
        # ancestors. Otherwise, one of the ancestors will have already inherited from
        # it.
        super_names.append("Class")
    else:
        super_names = [
            python_naming.class_name(inheritance.name)
            for inheritance in cls.inheritances
        ]

    assert len(super_names) > 0, "Assumption for the code generation below"

    writer = io.StringIO()
    if sum(len(super_name) + 2 for super_name in super_names) + len(name) < 70:
        # NOTE (mristin, 2022-09-17):
        # Put the class definition on the single line as it fit by the heuristic.
        super_names_joined = ", ".join(super_names)
        writer.write(f"class {name}({super_names_joined}):\n")
    else:
        writer.write(f"class {name}(\n")
        for i, super_name in enumerate(super_names):
            if i < len(super_names) - 1:
                writer.write(f"{II}{super_name},\n")
            else:
                writer.write(f"{II}{super_name}")
        writer.write("):\n")

    if docstring is not None:
        writer.write(textwrap.indent(docstring, I))
    else:
        writer.write(f"{I}# pylint: disable=missing-class-docstring")

    if len(blocks) == 0:
        # NOTE (mristin, 2022-09-28):
        # We have to add a ``pass`` statement if there was no description. Otherwise,
        # the generated code would be invalid. In cases where description is defined,
        # a ``pass`` statement is redundant.
        if docstring is None:
            writer.write(f"\n{I}pass")
    else:
        for i, code in enumerate(blocks):
            writer.write("\n\n")
            writer.write(textwrap.indent(code, I))

    # endregion

    return Stripped(writer.getvalue()), None


def _generate_abstract_visitor(symbol_table: intermediate.SymbolTable) -> Stripped:
    """Generate the code for the abstract visitor."""
    blocks = [
        Stripped(
            f"""\
def visit(
{II}self,
{II}that: Class
) -> None:
{I}\"\"\"Double-dispatch on :paramref:`that`.\"\"\"
{I}that.accept(self)"""
        )
    ]  # type: List[Stripped]

    for cls in symbol_table.concrete_classes:
        visit_name = python_naming.method_name(Identifier(f"visit_{cls.name}"))
        cls_name = python_naming.class_name(cls.name)

        blocks.append(
            Stripped(
                f"""\
@abc.abstractmethod
def {visit_name}(
{II}self,
{II}that: {cls_name}
) -> None:
{I}\"\"\"Visit :paramref:`that`.\"\"\"
{I}raise NotImplementedError()"""
            )
        )

    writer = io.StringIO()
    writer.write(
        f"""\
class AbstractVisitor:
{I}\"\"\"Visit the instances of the model.\"\"\"
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(textwrap.indent(block, I))

    return Stripped(writer.getvalue())


def _generate_abstract_visitor_with_context(
    symbol_table: intermediate.SymbolTable,
) -> Stripped:
    """Generate the code for the abstract visitor with context."""
    blocks = [
        Stripped(
            f"""\
def visit_with_context(
{II}self,
{II}that: Class,
{II}context: ContextT
) -> None:
{I}\"\"\"Double-dispatch on :paramref:`that`.\"\"\"
{I}that.accept_with_context(self, context)"""
        )
    ]  # type: List[Stripped]

    for cls in symbol_table.concrete_classes:
        visit_with_context_name = python_naming.method_name(
            Identifier(f"visit_{cls.name}_with_context")
        )
        cls_name = python_naming.class_name(cls.name)

        blocks.append(
            Stripped(
                f"""\
@abc.abstractmethod
def {visit_with_context_name}(
{II}self,
{II}that: {cls_name},
{II}context: ContextT
) -> None:
{I}\"\"\"Visit :paramref:`that` in :paramref:`context`.\"\"\"
{I}raise NotImplementedError()"""
            )
        )

    writer = io.StringIO()
    writer.write(
        f"""\
class AbstractVisitorWithContext(Generic[ContextT]):
{I}\"\"\"Visit the instances of the model with context.\"\"\"
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(textwrap.indent(block, I))

    return Stripped(writer.getvalue())


def _generate_pass_through_visitor(symbol_table: intermediate.SymbolTable) -> Stripped:
    """Generate the code for the pass-through visitor."""
    blocks = [
        Stripped(
            f"""\
def visit(
{II}self,
{II}that: Class
) -> None:
{I}\"\"\"Double-dispatch on :paramref:`that`.\"\"\"
{I}that.accept(self)"""
        )
    ]  # type: List[Stripped]

    for cls in symbol_table.concrete_classes:
        visit_name = python_naming.method_name(Identifier(f"visit_{cls.name}"))
        cls_name = python_naming.class_name(cls.name)

        blocks.append(
            Stripped(
                f"""\
def {visit_name}(
{II}self,
{II}that: {cls_name}
) -> None:
{I}\"\"\"Visit :paramref:`that`.\"\"\"
{I}for another in that.descend_once():
{II}self.visit(another)"""
            )
        )

    writer = io.StringIO()
    writer.write(
        f"""\
class PassThroughVisitor(AbstractVisitor):
{I}\"\"\"
{I}Visit the instances of the model without action.

{I}This visitor is not meant to be directly used. Instead, you usually
{I}inherit from it, and implement only the relevant visit methods.
{I}\"\"\"
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(textwrap.indent(block, I))

    return Stripped(writer.getvalue())


def _generate_pass_through_visitor_with_context(
    symbol_table: intermediate.SymbolTable,
) -> Stripped:
    """Generate the code for the pass-through visitor with context."""
    blocks = [
        Stripped(
            f"""\
def visit_with_context(
{II}self,
{II}that: Class,
{II}context: ContextT
) -> None:
{I}\"\"\"Double-dispatch on :paramref:`that`.\"\"\"
{I}that.accept_with_context(self, context)"""
        )
    ]  # type: List[Stripped]

    for cls in symbol_table.concrete_classes:
        visit_with_context_name = python_naming.method_name(
            Identifier(f"visit_{cls.name}_with_context")
        )
        cls_name = python_naming.class_name(cls.name)

        blocks.append(
            Stripped(
                f"""\
def {visit_with_context_name}(
{II}self,
{II}that: {cls_name},
{II}context: ContextT
) -> None:
{I}\"\"\"Visit :paramref:`that` in :paramref:`context`.\"\"\"
{I}for another in that.descend_once():
{II}self.visit_with_context(another, context)"""
            )
        )

    writer = io.StringIO()
    writer.write(
        f"""\
class PassThroughVisitorWithContext(
{II}AbstractVisitorWithContext[ContextT]
):
{I}\"\"\"
{I}Visit the instances of the model without action and in context.

{I}This visitor is not meant to be directly used. Instead, you usually
{I}inherit from it, and implement only the relevant visit methods.
{I}\"\"\"
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(textwrap.indent(block, I))

    return Stripped(writer.getvalue())


def _generate_abstract_transformer(symbol_table: intermediate.SymbolTable) -> Stripped:
    """Generate the code for the abstract transformer."""
    blocks = [
        Stripped(
            f"""\
def transform(
{II}self,
{II}that: Class
) -> T:
{I}\"\"\"Double-dispatch on :paramref:`that`.\"\"\"
{I}return that.transform(self)"""
        )
    ]  # type: List[Stripped]

    for cls in symbol_table.concrete_classes:
        transform_name = python_naming.method_name(Identifier(f"transform_{cls.name}"))

        cls_name = python_naming.class_name(cls.name)

        blocks.append(
            Stripped(
                f"""\
@abc.abstractmethod
def {transform_name}(
{II}self,
{II}that: {cls_name}
) -> T:
{I}\"\"\"Transform :paramref:`that`.\"\"\"
{I}raise NotImplementedError()"""
            )
        )

    writer = io.StringIO()
    writer.write(
        f"""\
class AbstractTransformer(Generic[T]):
{I}\"\"\"Transform the instances of the model.\"\"\"
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(textwrap.indent(block, I))

    return Stripped(writer.getvalue())


def _generate_abstract_transformer_with_context(
    symbol_table: intermediate.SymbolTable,
) -> Stripped:
    """Generate the code for the abstract transformer with context."""
    blocks = [
        Stripped(
            f"""\
def transform_with_context(
{II}self,
{II}that: Class,
{II}context: ContextT
) -> T:
{I}\"\"\"Double-dispatch on :paramref:`that`.\"\"\"
{I}return that.transform_with_context(self, context)"""
        )
    ]  # type: List[Stripped]

    for cls in symbol_table.concrete_classes:
        transform_with_context_name = python_naming.method_name(
            Identifier(f"transform_{cls.name}_with_context")
        )

        cls_name = python_naming.class_name(cls.name)

        blocks.append(
            Stripped(
                f"""\
@abc.abstractmethod
def {transform_with_context_name}(
{II}self,
{II}that: {cls_name},
{II}context: ContextT
) -> T:
{I}\"\"\"Transform :paramref:`that` in :paramref:`context`.\"\"\"
{I}raise NotImplementedError()"""
            )
        )

    writer = io.StringIO()
    writer.write(
        f"""\
class AbstractTransformerWithContext(
{II}Generic[ContextT, T]
):
{I}\"\"\"Transform the instances of the model in context.\"\"\"
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(textwrap.indent(block, I))

    return Stripped(writer.getvalue())


def _generate_transformer_with_default(
    symbol_table: intermediate.SymbolTable,
) -> Stripped:
    """Generate the code for the transformer with default transformation."""
    blocks = [
        Stripped(
            """\
#: Default value which is returned if no override of the transformation
default: T"""
        ),
        Stripped(
            f"""\
def __init__(self, default: T) -> None:
{I}\"\"\"Initialize with the given :paramref:`default` value.\"\"\"
{I}self.default = default"""
        ),
        Stripped(
            f"""\
def transform(
{II}self,
{II}that: Class
) -> T:
{I}\"\"\"Double-dispatch on :paramref:`that`.\"\"\"
{I}return that.transform(self)"""
        ),
    ]  # type: List[Stripped]

    for cls in symbol_table.concrete_classes:
        transform_name = python_naming.method_name(Identifier(f"transform_{cls.name}"))

        cls_name = python_naming.class_name(cls.name)

        blocks.append(
            Stripped(
                f"""\
def {transform_name}(
{II}self,
{II}that: {cls_name}
) -> T:
{I}\"\"\"Transform :paramref:`that`.\"\"\"
{I}return self.default"""
            )
        )

    writer = io.StringIO()
    writer.write(
        f"""\
class TransformerWithDefault(AbstractTransformer[T]):
{I}\"\"\"
{I}Transform the instances of the model.

{I}If you do not override the transformation methods, they simply
{I}return :py:attr:`.default`.
{I}\"\"\"
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(textwrap.indent(block, I))

    return Stripped(writer.getvalue())


def _generate_transformer_with_default_and_context(
    symbol_table: intermediate.SymbolTable,
) -> Stripped:
    """Generate the code for the transformer with default transformation and context."""
    blocks = [
        Stripped(
            """\
#: Default value which is returned if no override of the transformation
default: T"""
        ),
        Stripped(
            f"""\
def __init__(self, default: T) -> None:
{I}\"\"\"Initialize with the given :paramref:`default` value.\"\"\"
{I}self.default = default"""
        ),
        Stripped(
            f"""\
def transform_with_context(
{II}self,
{II}that: Class,
{II}context: ContextT
) -> T:
{I}\"\"\"Double-dispatch on :paramref:`that`.\"\"\"
{I}return that.transform_with_context(self, context)"""
        ),
    ]  # type: List[Stripped]

    for cls in symbol_table.concrete_classes:
        transform_with_context_name = python_naming.method_name(
            Identifier(f"transform_{cls.name}_with_context")
        )

        cls_name = python_naming.class_name(cls.name)

        blocks.append(
            Stripped(
                f"""\
def {transform_with_context_name}(
{II}self,
{II}that: {cls_name},
{II}context: ContextT
) -> T:
{I}\"\"\"Transform :paramref:`that` in :paramref:`context`.\"\"\"
{I}return self.default"""
            )
        )

    writer = io.StringIO()
    writer.write(
        f"""\
class TransformerWithDefaultAndContext(
{II}AbstractTransformerWithContext[ContextT, T]
):
{I}\"\"\"
{I}Transform the instances of the model in context.

{I}If you do not override the transformation methods, they simply
{I}return :py:attr:`.default`.
{I}\"\"\"
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(textwrap.indent(block, I))

    return Stripped(writer.getvalue())


def _generate_docstring_for_meta_model(
    description: intermediate.DescriptionOfMetaModel,
    qualified_module_name: python_common.QualifiedModuleName,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the docstring for the given meta-model."""
    text, errors = python_description.generate_summary_remarks_constraints(
        description=description,
        context=python_description.Context(
            qualified_module_name=qualified_module_name,
            module=Identifier("types"),
            cls_or_enum=None,
        ),
    )

    if errors is not None:
        return None, errors

    assert text is not None

    return python_description.docstring(text), None


# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
@ensure(
    lambda result:
    not (result[0] is not None) or result[0].endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    symbol_table: VerifiedIntermediateSymbolTable,
    qualified_module_name: python_common.QualifiedModuleName,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """
    Generate the Python code of the structures based on the symbol table.

    The ``qualified_module_name`` indicates the fully-qualified name of the base module.
    """
    errors = []  # type: List[Error]

    blocks = []  # type: List[Stripped]

    if symbol_table.meta_model.description is not None:
        # fmt: off
        docstring, docstring_errors = (
            _generate_docstring_for_meta_model(
                description=symbol_table.meta_model.description,
                qualified_module_name=qualified_module_name
            )
        )
        # fmt: on

        if docstring_errors is not None:
            errors.extend(docstring_errors)
        else:
            assert docstring is not None
            blocks.append(docstring)

    blocks.extend(
        [
            python_common.WARNING,
            Stripped(
                f"""\
import abc
import enum
from typing import (
{I}Generic,
{I}Iterator,
{I}Optional,
{I}TypeVar,
{I}List
)"""
            ),
            Stripped(
                """\
T = TypeVar("T")
ContextT = TypeVar("ContextT")"""
            ),
            Stripped(
                f"""\
class Class(abc.ABC):
{I}\"\"\"Represent the most general class of an AAS model.\"\"\"
{I}@abc.abstractmethod
{I}def descend_once(self) -> Iterator["Class"]:
{II}\"\"\"Iterate over all the instances referenced from this one.\"\"\"
{II}raise NotImplementedError()

{I}@abc.abstractmethod
{I}def descend(self) -> Iterator["Class"]:
{II}\"\"\"Iterate recursively over all the instances referenced from this one.\"\"\"
{II}raise NotImplementedError()

{I}@abc.abstractmethod
{I}def accept(
{III}self,
{III}visitor: "AbstractVisitor"
{I}) -> None:
{II}\"\"\"
{II}Dispatch the :paramref:`visitor` on this instance.

{II}:param visitor: to be dispatched
{II}\"\"\"
{II}raise NotImplementedError()

{I}@abc.abstractmethod
{I}def accept_with_context(
{III}self,
{III}visitor: "AbstractVisitorWithContext[ContextT]",
{III}context: ContextT
{I}) -> None:
{II}\"\"\"
{II}Dispatch the :paramref:`visitor` on this instance with :paramref:`context`.

{II}:param visitor: to be dispatched
{II}:param context: of the visitation
{II}\"\"\"
{II}raise NotImplementedError()

{I}@abc.abstractmethod
{I}def transform(
{III}self,
{III}transformer: "AbstractTransformer[T]"
{I}) -> T:
{II}\"\"\"
{II}Dispatch the :paramref:`transformer` on this instance.

{II}:param transformer: to be dispatched
{II}:return: transformed self
{II}\"\"\"
{II}raise NotImplementedError()

{I}@abc.abstractmethod
{I}def transform_with_context(
{III}self,
{III}transformer: "AbstractTransformerWithContext[ContextT, T]",
{III}context: ContextT
{I}) -> T:
{II}\"\"\"
{II}Dispatch the :paramref:`transformer` on this instance with :paramref:`context`.

{II}:param transformer: to be dispatched
{II}:return: transformed self
{II}\"\"\"
{II}raise NotImplementedError()"""
            ),
            Stripped(
                """\
# pylint: disable=redefined-builtin"""
            ),
        ]
    )

    for our_type in symbol_table.our_types:
        error: Optional[Error]

        if isinstance(our_type, intermediate.Enumeration):
            block, error = _generate_enum(
                enum=our_type, qualified_module_name=qualified_module_name
            )
            if error is None:
                assert block is not None
                blocks.append(block)
                continue

        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            # NOTE (mristin, 2022-09-28):
            # We do not generate the constrained primitives as types. We only
            # consider them in the verification.
            continue

        elif isinstance(
            our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            if our_type.is_implementation_specific:
                implementation_key = specific_implementations.ImplementationKey(
                    f"Types/{our_type.name}.py"
                )

                block = spec_impls.get(implementation_key, None)
                if block is None:
                    error = Error(
                        our_type.parsed.node,
                        f"The implementation is missing "
                        f"for the implementation-specific "
                        f"class: {implementation_key}",
                    )
                else:
                    blocks.append(block)
                    continue
            else:
                block, error = _generate_class(
                    cls=our_type,
                    spec_impls=spec_impls,
                    qualified_module_name=qualified_module_name,
                    symbol_table=symbol_table,
                )
                if error is None:
                    assert block is not None
                    blocks.append(block)
                    continue
        else:
            assert_never(our_type)

        assert error is not None
        errors.append(
            Error(
                our_type.parsed.node,
                f"Failed to generate the code for {our_type.name!r}",
                [error],
            )
        )

    if len(errors) > 0:
        return None, errors

    blocks.extend(
        [
            _generate_abstract_visitor(symbol_table=symbol_table),
            _generate_abstract_visitor_with_context(symbol_table=symbol_table),
            _generate_pass_through_visitor(symbol_table=symbol_table),
            _generate_pass_through_visitor_with_context(symbol_table=symbol_table),
            _generate_abstract_transformer(symbol_table=symbol_table),
            _generate_abstract_transformer_with_context(symbol_table=symbol_table),
            _generate_transformer_with_default(symbol_table=symbol_table),
            _generate_transformer_with_default_and_context(symbol_table=symbol_table),
        ]
    )

    blocks.append(python_common.WARNING)

    out = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            out.write("\n\n\n")

        out.write(block)

    out.write("\n")

    return out.getvalue(), None


# endregion



"""Generate C++ constants corresponding to the constants of the meta-model."""

from aas_core_codegen.cpp.constants import _generate

generate_header = _generate.generate_header
generate_implementation = _generate.generate_implementation



"""Provide common functions shared among different Java code generation modules."""
import re
from typing import List, cast, Optional

from icontract import ensure, require

from aas_core_codegen import intermediate
from aas_core_codegen.common import Stripped, assert_never
from aas_core_codegen.java import naming as java_naming


@ensure(lambda result: result.startswith('"'))
@ensure(lambda result: result.endswith('"'))
def string_literal(text: str) -> Stripped:
    """Generate a Java string literal from the ``text``."""
    escaped = []  # type: List[str]

    for character in text:
        if character == "\t":
            escaped.append("\\t")
        elif character == "\b":
            escaped.append("\\b")
        elif character == "\n":
            escaped.append("\\n")
        elif character == "\r":
            escaped.append("\\r")
        elif character == "\f":
            escaped.append("\\f")
        elif character == "'":
            escaped.append("\\'")
        elif character == '"':
            escaped.append('\\"')
        elif character == "\\":
            escaped.append("\\\\")
        else:
            escaped.append(character)

    return Stripped('"{}"'.format("".join(escaped)))


def needs_escaping(text: str) -> bool:
    """Check whether the ``text`` contains a character that needs escaping."""
    for character in text:
        if character == "\t":
            return True
        elif character == "\b":
            return True
        elif character == "\n":
            return True
        elif character == "\r":
            return True
        elif character == "\f":
            return True
        elif character == "'":
            return True
        elif character == '"':
            return True
        elif character == "\\":
            return True
        else:
            pass

    return False


PRIMITIVE_TYPE_MAP = {
    intermediate.PrimitiveType.BOOL: Stripped("Boolean"),
    intermediate.PrimitiveType.INT: Stripped("Long"),
    intermediate.PrimitiveType.FLOAT: Stripped("Float"),
    intermediate.PrimitiveType.STR: Stripped("String"),
    intermediate.PrimitiveType.BYTEARRAY: Stripped("byte[]"),
}


# fmt: off
@require(
    lambda our_type_qualifier:
    not (our_type_qualifier is not None)
    or not our_type_qualifier.endswith('.')
)
# fmt: on
def generate_type(
    type_annotation: intermediate.TypeAnnotationUnion,
    our_type_qualifier: Optional[Stripped] = None,
) -> Stripped:
    """
    Generate the Java type for the given type annotation.

    ``our_type_prefix`` is appended to all our types, if specified.
    """
    our_type_prefix = "" if our_type_qualifier is None else f"{our_type_qualifier}."
    # BEFORE-RELEASE (empwilli, 2023-12-14): test in isolation
    if isinstance(type_annotation, intermediate.PrimitiveTypeAnnotation):
        return PRIMITIVE_TYPE_MAP[type_annotation.a_type]

    elif isinstance(type_annotation, intermediate.OurTypeAnnotation):
        our_type = type_annotation.our_type

        if isinstance(our_type, intermediate.Enumeration):
            return Stripped(
                our_type_prefix + java_naming.enum_name(type_annotation.our_type.name)
            )

        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            return PRIMITIVE_TYPE_MAP[our_type.constrainee]

        elif isinstance(our_type, intermediate.Class):
            # NOTE (empwilli, 2023-12-14):
            # We want to allow custom enhancements and wrappings around
            # our model classes. Therefore, we always operate over Java interfaces
            # instead of concrete classes, even if the class is a concrete one and
            # has no concrete descendants.

            return Stripped(our_type_prefix + java_naming.interface_name(our_type.name))

    elif isinstance(type_annotation, intermediate.ListTypeAnnotation):
        item_type = generate_type(
            type_annotation=type_annotation.items, our_type_qualifier=our_type_qualifier
        )

        return Stripped(f"List<{item_type}>")

    elif isinstance(type_annotation, intermediate.OptionalTypeAnnotation):
        value = generate_type(
            type_annotation=type_annotation.value, our_type_qualifier=our_type_qualifier
        )
        return Stripped(f"Optional<{value}>")

    else:
        assert_never(type_annotation)

    raise AssertionError("Should not have gotten here")


INDENT = "  "
INDENT2 = INDENT * 2
INDENT3 = INDENT * 3
INDENT4 = INDENT * 4
INDENT5 = INDENT * 5
INDENT6 = INDENT * 6


INTERFACE_PKG = "model"
CLASS_PKG = "impl"
ENUM_PKG = "enums"


def interface_package_path(name: Stripped) -> Stripped:
    """Create the package path for an interface file."""
    return Stripped(f"{INTERFACE_PKG}/{name}.java")


def class_package_path(name: Stripped) -> Stripped:
    """Create the package path for an interface file."""
    return Stripped(f"{CLASS_PKG}/{name}.java")


def enum_package_path(name: Stripped) -> Stripped:
    """Create the package path for an interface file."""
    return Stripped(f"{ENUM_PKG}/{name}.java")


# noinspection RegExpSimplifiable
PACKAGE_IDENTIFIER_RE = re.compile(r"[a-z][a-z_0-9]*(\.[a-z][a-z_0-9]*)*")


class PackageIdentifier(str):
    """Capture a package identifier."""

    @require(lambda identifier: PACKAGE_IDENTIFIER_RE.fullmatch(identifier))
    def __new__(cls, identifier: str) -> "PackageIdentifier":
        return cast(PackageIdentifier, identifier)


WARNING = Stripped(
    """\
/*
 * This code has been automatically generated by aas-core-codegen.
 * Do NOT edit or append.
 */"""
)


class JavaFile:
    """Representation of a Java source file."""

    # fmt: off
    @require(lambda name, content: (len(name) > 0) and (len(content) > 0))
    @require(lambda content: content.endswith('\n'), "Trailing newline mandatory for valid end-of-files")
    # fmt: on
    def __init__(
        self,
        name: str,
        content: str,
    ):
        self.name = name
        self.content = content



"""Generate C++ code to de/wstringify enumerations and primitives."""

import io
from typing import List

from icontract import ensure

import aas_core_codegen.naming
from aas_core_codegen import intermediate
from aas_core_codegen.common import Stripped, Identifier, indent_but_first_line
from aas_core_codegen.cpp import (
    common as cpp_common,
    naming as cpp_naming,
)
from aas_core_codegen.cpp.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
    INDENT5 as IIIII,
)


def _generate_model_type_from_wstring_definition() -> List[Stripped]:
    """Generate the definition of translation of a model type from wstring."""
    enum_name = cpp_naming.enum_name(Identifier("Model_type"))
    from_wstring = cpp_naming.function_name(Identifier("model_type_from_wstring"))
    must_from_wstring = cpp_naming.function_name(
        Identifier("must_model_type_from_wstring")
    )

    return [
        Stripped(
            f"""\
/**
 * Try to parse the \\p text as a model type literal.
 *
 * \\param text to be parsed
 * \\return literal, or nothing, if \\p text invalid
 */
common::optional<types::{enum_name}> {from_wstring}(
{I}const std::wstring& text
);"""
        ),
        Stripped(
            f"""\
/**
 * Parse the \\p text as a model type literal.
 *
 * \\param text to be parsed
 * \\return literal
 * \\throw std::invalid_argument if \\p text invalid
 */
types::{enum_name} {must_from_wstring}(
{I}const std::wstring& text
);"""
        ),
    ]


def _generate_model_type_from_wstring_implementation(
    symbol_table: intermediate.SymbolTable,
) -> List[Stripped]:
    """Generate the implementation of translation of a model type from wstring."""
    blocks = []  # type: List[Stripped]

    # region Map

    enum_name = cpp_naming.enum_name(Identifier("Model_type"))

    items = []  # type: List[str]
    for cls in symbol_table.concrete_classes:
        literal_name = cpp_naming.enum_literal_name(cls.name)

        literal_value = cpp_common.wstring_literal(
            aas_core_codegen.naming.json_model_type(cls.name)
        )

        items.append(
            Stripped(
                f"""\
{{
{I}{literal_value},
{I}types::{enum_name}::{literal_name}
}}"""
            )
        )

    map_name = cpp_naming.constant_name(Identifier("model_type_from_wstring_map"))

    items_joined = ",\n".join(items)

    blocks.append(
        Stripped(
            f"""\
const std::unordered_map<
{I}std::wstring,
{I}types::{enum_name}
> {map_name} = {{
{I}{indent_but_first_line(items_joined, I)}
}};"""
        )
    )

    # endregion

    # region From function

    from_wstring = cpp_naming.function_name(Identifier("model_type_from_wstring"))

    blocks.append(
        Stripped(
            f"""\
common::optional<types::{enum_name}> {from_wstring}(
{I}const std::wstring& text
) {{
{I}const auto it = {map_name}.find(
{II}text
{I});
{I}if (it == {map_name}.end()) {{
{II}return {{}};
{I}}}
{I}return it->second;
}}"""
        )
    )

    # endregion

    # region Must-from function

    must_from_wstring = cpp_naming.function_name(
        Identifier("must_model_type_from_wstring")
    )

    blocks.append(
        Stripped(
            f"""\
types::{enum_name} {must_from_wstring}(
{I}const std::wstring& text
) {{
{I}const auto it = {map_name}.find(
{II}text
{I});
{I}if (it == {map_name}.end()) {{
{II}throw std::invalid_argument(
{III}common::WstringToUtf8(
{IIII}common::Concat(
{IIIII}L"Unexpected {enum_name} literal: ",
{IIIII}text
{IIII})
{III})
{II});
{I}}}
{I}return it->second;
}}"""
        )
    )

    # endregion

    return blocks


def _generate_model_type_to_wstring_definition() -> Stripped:
    """Generate the definition for translating a wstring to a model type."""
    enum_name = cpp_naming.enum_name(Identifier("Model_type"))

    model_type_arg = cpp_naming.argument_name(Identifier("model_type"))

    # NOTE (mristin, 2023-07-12):
    # We use ``to_wstring`` naming to resemble the functions in standard library.
    return Stripped(
        f"""\
/**
 * Translate the enumeration literal \\p {model_type_arg} to text.
 *
 * \\param {model_type_arg} to be converted into text
 * \\return text representation of \\p {model_type_arg}
 * \\throw std::invalid_argument if \\p {model_type_arg} invalid
 */
std::wstring to_wstring(
{I}types::{enum_name} {model_type_arg}
);"""
    )


def _generate_model_type_to_wstring_implementation(
    symbol_table: intermediate.SymbolTable,
) -> Stripped:
    """Generate the implementation for translating a wstring to a model type."""
    enum_name = cpp_naming.enum_name(Identifier("Model_type"))

    case_blocks = []  # type: List[Stripped]
    for cls in symbol_table.concrete_classes:
        literal_name = cpp_naming.enum_literal_name(cls.name)
        literal_value = cpp_common.wstring_literal(
            aas_core_codegen.naming.json_model_type(cls.name)
        )

        case_blocks.append(
            Stripped(
                f"""\
case types::{enum_name}::{literal_name}:
return {literal_value};"""
            )
        )

    model_type_arg = cpp_naming.argument_name(Identifier("model_type"))

    case_blocks.append(
        Stripped(
            f"""\
default:
{I}throw std::invalid_argument(
{II}common::Concat(
{III}"Unexpected {enum_name} literal: ",
{III}std::to_string(
{IIII}static_cast<std::uint32_t>({model_type_arg})
{III})
{II})
{I});"""
        )
    )

    case_blocks_joined = "\n".join(case_blocks)

    # NOTE (mristin, 2023-07-12):
    # We use ``to_wstring`` naming to resemble the functions in standard library.
    return Stripped(
        f"""\
std::wstring to_wstring(
{I}types::{enum_name} {model_type_arg}
) {{
{I}switch ({model_type_arg}) {{
{II}{indent_but_first_line(case_blocks_joined, II)}
{I}}}
}}"""
    )


def _generate_enum_from_wstring_definition(
    enum: intermediate.Enumeration,
) -> List[Stripped]:
    """Generate the definition of translation of an enum from wstring."""
    enum_name = cpp_naming.enum_name(enum.name)
    from_wstring = cpp_naming.function_name(Identifier(f"{enum.name}_from_wstring"))
    must_from_wstring = cpp_naming.function_name(
        Identifier(f"must_{enum.name}_from_wstring")
    )

    return [
        Stripped(
            f"""\
/**
 * Try to parse the \\p text as a literal of
 * types::{enum_name}.
 *
 * \\param text to be parsed
 * \\return literal, or nothing, if \\p text invalid
 */
common::optional<types::{enum_name}> {from_wstring}(
{I}const std::wstring& text
);"""
        ),
        Stripped(
            f"""\
/**
 * Parse the \\p text as a literal of
 * types::{enum_name}.
 *
 * \\param text to be parsed
 * \\return literal
 * \\throw std::invalid_argument if \\p text invalid
 */
types::{enum_name} {must_from_wstring}(
{I}const std::wstring& text
);"""
        ),
    ]


def _generate_enum_from_wstring_implementation(
    enum: intermediate.Enumeration,
) -> List[Stripped]:
    """Generate the implementation of translation of an  enum from wstring."""
    blocks = []  # type: List[Stripped]

    # region Map

    enum_name = cpp_naming.enum_name(enum.name)

    items = []  # type: List[str]
    for literal in enum.literals:
        literal_name = cpp_naming.enum_literal_name(literal.name)
        literal_value = cpp_common.wstring_literal(literal.value)

        items.append(
            Stripped(
                f"""\
{{
{I}{literal_value},
{I}types::{enum_name}::{literal_name}
}}"""
            )
        )

    map_name = cpp_naming.constant_name(Identifier(f"{enum.name}_from_wstring_map"))

    items_joined = ",\n".join(items)

    blocks.append(
        Stripped(
            f"""\
const std::unordered_map<
{I}std::wstring,
{I}types::{enum_name}
> {map_name} = {{
{I}{indent_but_first_line(items_joined, I)}
}};"""
        )
    )

    # endregion

    # region From function

    from_wstring = cpp_naming.function_name(Identifier(f"{enum.name}_from_wstring"))

    blocks.append(
        Stripped(
            f"""\
common::optional<types::{enum_name}> {from_wstring}(
{I}const std::wstring& text
) {{
{I}const auto it = {map_name}.find(
{II}text
{I});
{I}if (it == {map_name}.end()) {{
{II}return {{}};
{I}}}
{I}return it->second;
}}"""
        )
    )

    # endregion

    # region Must-from function

    must_from_wstring = cpp_naming.function_name(
        Identifier(f"must_{enum.name}_from_wstring")
    )

    blocks.append(
        Stripped(
            f"""\
types::{enum_name} {must_from_wstring}(
{I}const std::wstring& text
) {{
{I}const auto it = {map_name}.find(
{II}text
{I});
{I}if (it == {map_name}.end()) {{
{II}throw std::invalid_argument(
{III}common::WstringToUtf8(
{IIII}common::Concat(
{IIIII}L"Unexpected {enum_name} literal: ",
{IIIII}text
{IIII})
{III})
{II});
{I}}}
{I}return it->second;
}}"""
        )
    )

    # endregion

    return blocks


def _generate_enum_to_wstring_definition(enum: intermediate.Enumeration) -> Stripped:
    """Generate the definition for translating a wstring to an enum literal."""
    enum_name = cpp_naming.enum_name(enum.name)

    literal_arg = cpp_naming.argument_name(Identifier("literal"))

    # NOTE (mristin, 2023-07-12):
    # We use ``to_wstring`` naming to resemble the functions in standard library.
    return Stripped(
        f"""\
/**
 * Translate the enumeration literal \\p {literal_arg}
 * of types::{enum_name} to text.
 *
 * \\param {literal_arg} to be converted into text
 * \\return text representation of \\p {literal_arg}
 * \\throw std::invalid_argument if \\p {literal_arg} invalid
 */
std::wstring to_wstring(
{I}types::{enum_name} {literal_arg}
);"""
    )


def _generate_enum_to_wstring_implementation(
    enum: intermediate.Enumeration,
) -> Stripped:
    """Generate the implementation for translating a wstring to an enum literal."""
    enum_name = cpp_naming.enum_name(enum.name)

    case_blocks = []  # type: List[Stripped]
    for literal in enum.literals:
        literal_name = cpp_naming.enum_literal_name(literal.name)
        literal_value = cpp_common.wstring_literal(literal.value)

        case_blocks.append(
            Stripped(
                f"""\
case types::{enum_name}::{literal_name}:
{I}return {literal_value};"""
            )
        )

    literal_arg = cpp_naming.argument_name(Identifier("literal"))

    case_blocks.append(
        Stripped(
            f"""\
default:
{I}throw std::invalid_argument(
{II}common::Concat(
{III}"Unexpected {enum_name} literal: ",
{III}std::to_string(
{IIII}static_cast<std::uint32_t>({literal_arg})
{III})
{II})
{I});"""
        )
    )

    case_blocks_joined = "\n".join(case_blocks)

    # NOTE (mristin, 2023-07-12):
    # We use ``to_wstring`` naming to resemble the functions in standard library.
    return Stripped(
        f"""\
std::wstring to_wstring(
{I}types::{enum_name} {literal_arg}
) {{
{I}switch ({literal_arg}) {{
{II}{indent_but_first_line(case_blocks_joined, II)}
{I}}}
}}"""
    )


# fmt: off
@ensure(
    lambda result:
    result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate_header(
    symbol_table: intermediate.SymbolTable, library_namespace: Stripped
) -> str:
    """Generate the C++ header code for stringification."""
    namespace = Stripped(f"{library_namespace}::wstringification")

    include_guard_var = cpp_common.include_guard_var(namespace)

    include_prefix_path = cpp_common.generate_include_prefix_path(library_namespace)

    blocks = [
        Stripped(
            f"""\
#ifndef {include_guard_var}
#define {include_guard_var}"""
        ),
        cpp_common.WARNING,
        Stripped(
            f"""\
#include "{include_prefix_path}/common.hpp"
#include "{include_prefix_path}/types.hpp"

#pragma warning(push, 0)
#include <string>
#include <vector>
#pragma warning(pop)"""
        ),
        cpp_common.generate_namespace_opening(library_namespace),
        Stripped(
            """\
/**
 * \\defgroup wstringification De/wstringify to and from wstrings (where applicable).
 * @{
 */
namespace wstringification {"""
        ),
        *_generate_model_type_from_wstring_definition(),
        _generate_model_type_to_wstring_definition(),
    ]

    for enum in symbol_table.enumerations:
        blocks.extend(_generate_enum_from_wstring_definition(enum=enum))
        blocks.append(_generate_enum_to_wstring_definition(enum=enum))

    blocks.extend(
        [
            Stripped(
                """\
}  // namespace wstringification
/**@}*/"""
            ),
            cpp_common.generate_namespace_closing(library_namespace),
            cpp_common.WARNING,
            Stripped(f"#endif  // {include_guard_var}"),
        ]
    )

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


# fmt: off
@ensure(
    lambda result:
    result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate_implementation(
    symbol_table: intermediate.SymbolTable, library_namespace: Stripped
) -> str:
    """Generate the C++ code for stringification."""
    namespace = Stripped(f"{library_namespace}::wstringification")

    include_prefix_path = cpp_common.generate_include_prefix_path(library_namespace)

    blocks = [
        cpp_common.WARNING,
        Stripped(
            f"""\
#include "{include_prefix_path}/wstringification.hpp"

#pragma warning(push, 0)
#include <unordered_map>
#pragma warning(pop)"""
        ),
        cpp_common.generate_namespace_opening(namespace),
        *_generate_model_type_from_wstring_implementation(symbol_table=symbol_table),
        _generate_model_type_to_wstring_implementation(symbol_table=symbol_table),
    ]  # type: List[Stripped]

    for enum in symbol_table.enumerations:
        blocks.extend(_generate_enum_from_wstring_implementation(enum=enum))
        blocks.append(_generate_enum_to_wstring_implementation(enum=enum))

    blocks.extend(
        [
            cpp_common.generate_namespace_closing(namespace),
            cpp_common.WARNING,
        ]
    )

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()



"""Generate the Python constants corresponding to the constants of the meta-model."""
import io
import textwrap
from typing import (
    Optional,
    List,
    Tuple,
)

from icontract import ensure

from aas_core_codegen import intermediate
from aas_core_codegen.common import (
    Error,
    assert_never,
    Stripped,
    indent_but_first_line,
    Identifier,
)
from aas_core_codegen.python import (
    common as python_common,
    naming as python_naming,
    description as python_description,
)
from aas_core_codegen.python.common import (
    INDENT as I,
)


# region Generation


def _generate_documentation_comment_for_constant(
    description: intermediate.DescriptionOfConstant, context: python_description.Context
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the docstring for the given constant."""
    text, errors = python_description.generate_summary_remarks(
        description=description, context=context
    )

    if errors is not None:
        return None, errors

    assert text is not None

    return python_description.documentation_comment(text), None


@ensure(lambda result: (result[0] is None) ^ (result[1] is None))
def _generate_constant_primitive(
    constant: intermediate.ConstantPrimitive,
    qualified_module_name: python_common.QualifiedModuleName,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the definition of a constant primitive."""
    writer = io.StringIO()

    if constant.description is not None:
        docstring, docstring_errors = _generate_documentation_comment_for_constant(
            description=constant.description,
            context=python_description.Context(
                qualified_module_name=qualified_module_name,
                module=Identifier("constants"),
                cls_or_enum=None,
            ),
        )
        if docstring_errors is not None:
            return None, Error(
                constant.parsed.node,
                f"Failed to generate the documentation comment for {constant.name!r}",
                docstring_errors,
            )

        assert docstring is not None
        writer.write(docstring)
        writer.write("\n")

    constant_name = python_naming.constant_name(constant.name)

    if constant.a_type is intermediate.PrimitiveType.BOOL:
        literal = "True" if constant.value else "False"

        writer.write(f"{constant_name}: bool = {literal}")

    elif constant.a_type is intermediate.PrimitiveType.INT:
        literal = str(constant.value)

        writer.write(f"{constant_name}: int = {literal}")

    elif constant.a_type is intermediate.PrimitiveType.FLOAT:
        # NOTE (mristin, 2022-09-28):
        # We assume that the float constants are not meant to be all to precise.
        # Therefore, we use a string representation here. However, beware that we
        # might have to use a more precise representation in the future if the spec
        # change.
        literal = str(constant.value)

        writer.write(f"{constant_name}: float = {literal}")

    elif constant.a_type is intermediate.PrimitiveType.STR:
        assert isinstance(constant.value, str)
        literal = python_common.string_literal(constant.value)

        writer.write(f"{constant_name}: str = {literal}")

    elif constant.a_type is intermediate.PrimitiveType.BYTEARRAY:
        assert isinstance(constant.value, bytearray)

        literal, _ = python_common.bytes_literal(value=constant.value)

        writer.write(
            f"""\
{constant_name}: bytes = (
{I}{indent_but_first_line(literal, I)}
)"""
        )

    else:
        assert_never(constant.a_type)

    return Stripped(writer.getvalue()), None


@ensure(lambda result: (result[0] is None) ^ (result[1] is None))
def _generate_constant_set_of_primitives(
    constant: intermediate.ConstantSetOfPrimitives,
    qualified_module_name: python_common.QualifiedModuleName,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the definition of a constant set of primitives."""
    writer = io.StringIO()

    if constant.description is not None:
        docstring, docstring_errors = _generate_documentation_comment_for_constant(
            description=constant.description,
            context=python_description.Context(
                qualified_module_name=qualified_module_name,
                module=Identifier("constants"),
                cls_or_enum=None,
            ),
        )
        if docstring_errors is not None:
            return None, Error(
                constant.parsed.node,
                f"Failed to generate the documentation comment for {constant.name!r}",
                docstring_errors,
            )

        assert docstring is not None
        writer.write(docstring)
        writer.write("\n")

    constant_name = python_naming.constant_name(constant.name)

    if constant.a_type is intermediate.PrimitiveType.BOOL:
        writer.write(
            f"""\
{constant_name}: Set[bool] = {{
"""
        )

        for i, literal in enumerate(constant.literals):
            writer.write(textwrap.indent("True" if literal.value else "False", I))

            if i < len(constant.literals) - 1:
                writer.write(",\n")
            else:
                writer.write("\n")

        writer.write("}")

    elif constant.a_type is intermediate.PrimitiveType.INT:
        writer.write(
            f"""\
{constant_name}: Set[int] = {{
"""
        )

        for i, literal in enumerate(constant.literals):
            writer.write(textwrap.indent(str(literal.value), I))

            if i < len(constant.literals) - 1:
                writer.write(",\n")
            else:
                writer.write("\n")

        writer.write("}")

    elif constant.a_type is intermediate.PrimitiveType.FLOAT:
        # NOTE (mristin, 2022-07-06):
        # We assume that the float constants are not meant to be all to precise.
        # Therefore, we use a string representation here. However, beware that we
        # might have to use a more precise representation in the future if the spec
        # change.

        writer.write(
            f"""\
{constant_name}: Set[float] = {{
"""
        )

        for i, literal in enumerate(constant.literals):
            writer.write(textwrap.indent(str(literal.value), I))

            if i < len(constant.literals) - 1:
                writer.write(",\n")
            else:
                writer.write("\n")

        writer.write("}")

    elif constant.a_type is intermediate.PrimitiveType.STR:
        writer.write(
            f"""\
{constant_name}: Set[str] = {{
"""
        )

        for i, literal in enumerate(constant.literals):
            assert isinstance(literal.value, str)

            writer.write(
                textwrap.indent(python_common.string_literal(literal.value), I)
            )

            if i < len(constant.literals) - 1:
                writer.write(",\n")
            else:
                writer.write("\n")

        writer.write("}")

    elif constant.a_type is intermediate.PrimitiveType.BYTEARRAY:
        writer.write(
            f"""\
{constant_name}: Set[bytes] = {{
"""
        )

        for i, literal in enumerate(constant.literals):
            assert isinstance(literal.value, bytearray)

            literal_in_code, _ = python_common.bytes_literal(bytes(literal.value))

            writer.write(textwrap.indent(literal_in_code, I))

            if i < len(constant.literals) - 1:
                writer.write(",\n")
            else:
                writer.write("\n")

        writer.write("}")

    else:
        assert_never(constant.a_type)

    return Stripped(writer.getvalue()), None


@ensure(lambda result: (result[0] is None) ^ (result[1] is None))
def _generate_constant_set_of_enumeration_literals(
    constant: intermediate.ConstantSetOfEnumerationLiterals,
    qualified_module_name: python_common.QualifiedModuleName,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the definition of a constant set of enumeration literals."""
    writer = io.StringIO()

    if constant.description is not None:
        docstring, docstring_errors = _generate_documentation_comment_for_constant(
            description=constant.description,
            context=python_description.Context(
                qualified_module_name=qualified_module_name,
                module=Identifier("constants"),
                cls_or_enum=None,
            ),
        )
        if docstring_errors is not None:
            return None, Error(
                constant.parsed.node,
                f"Failed to generate the documentation comment for {constant.name!r}",
                docstring_errors,
            )

        assert docstring is not None
        writer.write(docstring)
        writer.write("\n")

    constant_name = python_naming.constant_name(constant.name)
    enum_name = python_naming.enum_name(constant.enumeration.name)

    writer.write(
        f"""\
{constant_name}: Set[aas_types.{enum_name}] = {{
"""
    )

    for i, literal in enumerate(constant.literals):
        literal_name = python_naming.enum_literal_name(literal.name)

        writer.write(textwrap.indent(f"aas_types.{enum_name}.{literal_name}", I))

        if i < len(constant.literals) - 1:
            writer.write(",\n")
        else:
            writer.write("\n")

    writer.write("}")

    return Stripped(writer.getvalue()), None


# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
@ensure(
    lambda result:
    not (result[0] is not None) or result[0].endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    symbol_table: intermediate.SymbolTable,
    qualified_module_name: python_common.QualifiedModuleName,
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """
    Generate the Python code of the constants based on the symbol table.

    The ``qualified_module_name`` indicates the fully-qualified name of the base module.
    """
    errors = []  # type: List[Error]

    blocks = [
        Stripped('"""Provide constant values of the meta-model."""'),
        python_common.WARNING,
        Stripped(
            f"""\
from typing import Set

import {qualified_module_name}.types as aas_types"""
        ),
    ]  # type: List[Stripped]

    for constant in symbol_table.constants:
        block: Optional[Stripped]
        error: Optional[Error]

        if isinstance(constant, intermediate.ConstantPrimitive):
            block, error = _generate_constant_primitive(
                constant=constant, qualified_module_name=qualified_module_name
            )
        elif isinstance(constant, intermediate.ConstantSetOfPrimitives):
            block, error = _generate_constant_set_of_primitives(
                constant=constant, qualified_module_name=qualified_module_name
            )
        elif isinstance(constant, intermediate.ConstantSetOfEnumerationLiterals):
            block, error = _generate_constant_set_of_enumeration_literals(
                constant=constant, qualified_module_name=qualified_module_name
            )
        else:
            assert_never(constant)

        if error is not None:
            errors.append(error)
            continue

        assert block is not None
        blocks.append(block)

    if len(errors) > 0:
        return None, errors

    blocks.append(python_common.WARNING)

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue(), None


# endregion



"""Generate the test code for the JSON de/serialization of enums."""
import json
from typing import List

from icontract import ensure

from aas_core_codegen import intermediate
from aas_core_codegen.common import Stripped, indent_but_first_line
from aas_core_codegen.csharp import common as csharp_common, naming as csharp_naming
from aas_core_codegen.csharp.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
)


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    namespace: csharp_common.NamespaceIdentifier,
    symbol_table: intermediate.SymbolTable,
) -> str:
    """
    Generate the test code for the JSON de/serialization of enums.

    The ``namespace`` indicates the fully-qualified name of the base project.
    """
    blocks = []  # type: List[Stripped]

    for enumeration in symbol_table.enumerations:
        enum_name = csharp_naming.enum_name(enumeration.name)

        assert (
            len(enumeration.literals) > 0
        ), f"Unexpected enumeration without literals: {enumeration.name}"

        literal_value = enumeration.literals[0].value
        literal_value_json_str = json.dumps(literal_value)

        blocks.append(
            Stripped(
                f"""\
[Test]
public void Test_round_trip_{enum_name}()
{{
{I}var node = Nodes.JsonValue.Create(
{II}{csharp_common.string_literal(literal_value)})
{III}?? throw new System.InvalidOperationException(
{IIII}"Unexpected null node");

{I}var parsed = Aas.Jsonization.Deserialize.{enum_name}From(
{II}node);

{I}var serialized = Aas.Jsonization.Serialize.{enum_name}ToJsonValue(
{II}parsed);

{I}Assert.AreEqual(
{II}{csharp_common.string_literal(literal_value_json_str)},
{II}serialized.ToJsonString());
}}  // void Test_round_trip_{enum_name}"""
            )
        )

    blocks_joined = "\n\n".join(blocks)

    return f"""\
{csharp_common.WARNING}

using Aas = {namespace};  // renamed

using Nodes = System.Text.Json.Nodes;

using NUnit.Framework;  // can't alias

namespace {namespace}.Tests
{{
{I}public class TestJsonizationOfEnums
{I}{{
{II}{indent_but_first_line(blocks_joined, II)}
{I}}}  // class TestJsonizationOfEnums
}}  // namespace {namespace}.Tests

{csharp_common.WARNING}
"""


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate the test code for the ``OverXOrEmpty`` methods."""

from typing import List

from icontract import ensure

from aas_core_codegen import intermediate
from aas_core_codegen.common import Stripped, indent_but_first_line, Identifier
from aas_core_codegen.csharp import common as csharp_common, naming as csharp_naming
from aas_core_codegen.csharp.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
)


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    namespace: csharp_common.NamespaceIdentifier,
    symbol_table: intermediate.SymbolTable,
) -> str:
    """
    Generate the test code for the ``OverXOrEmpty`` methods.

    The ``namespace`` indicates the fully-qualified name of the base project.
    """
    blocks = []  # type: List[Stripped]

    for concrete_cls in symbol_table.concrete_classes:
        cls_name_csharp = csharp_naming.class_name(concrete_cls.name)

        for prop in concrete_cls.properties:
            method_name_csharp = csharp_naming.method_name(
                Identifier(f"Over_{prop.name}_or_empty")
            )

            prop_name_csharp = csharp_naming.property_name(prop.name)

            if isinstance(
                prop.type_annotation, intermediate.OptionalTypeAnnotation
            ) and isinstance(
                prop.type_annotation.value, intermediate.ListTypeAnnotation
            ):
                blocks.append(
                    Stripped(
                        f"""\
[Test]
public void Test_{cls_name_csharp}_{method_name_csharp}()
{{
{I}foreach (Aas.{cls_name_csharp} instance in new[]
{I}{{
{II}Aas.Tests.CommonJsonization.LoadMinimal{cls_name_csharp}(),
{II}Aas.Tests.CommonJsonization.LoadMaximal{cls_name_csharp}()
{I}}})
{I}{{
{II}int count = 0;
{II}foreach (var _ in instance.{method_name_csharp}())
{II}{{
{III}count++;
{II}}}

{II}Assert.AreEqual(
{III}instance.{prop_name_csharp}?.Count ?? 0,
{III}count);
{I}}}
}}  // public void Test_{cls_name_csharp}_{method_name_csharp}"""
                    )
                )

    blocks_joined = "\n\n".join(blocks)

    return f"""\
{csharp_common.WARNING}

using Aas = {namespace};  // renamed

using NUnit.Framework;  // can't alias

namespace {namespace}.Tests
{{
{I}public class TestOverXOrEmpty
{I}{{
{II}{indent_but_first_line(blocks_joined, II)}
{I}}}  // class TestOverXOrEmpty
}}  // namespace {namespace}.Tests

{csharp_common.WARNING}
"""


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())
